@@ -149,6 +149,11 @@ sub get_deps {
   } split (/\0/, $self->{deps});
 }
 
+sub set_deps {
+  my ($self, $deps) = @_;
+  $self->{deps}	= $self->mk_deps ($deps);
+}
+
 sub mk_deps {
   my ($self, $deps) = @_;
   my @compressed = ();
@@ -214,6 +219,10 @@ The metadatum is converted to its native type, e.g. C<score> is return as an
 integer, C<title> as a string, etc.  If the metadatum is not provided, the
 default value for that item, defined in HTML::WebMake::Metadata, is used.
 
+Note that this method should only be called from a deferred reference, as
+metadata often isn't available until all the normal content references in the
+current page have been expanded.
+
 =cut
 
 sub get_metadata {
@@ -275,6 +284,20 @@ sub get_modtime {
 
 # -------------------------------------------------------------------------
 
+=item $url = $cont->get_edit_href();
+
+Return the URL used to edit a content item, using the WebMake CGI
+edit script.
+
+=cut
+
+sub get_edit_href {
+  my ($self) = @_;
+  return '';
+}
+
+# -------------------------------------------------------------------------
+
 sub set_declared {
   my ($self, $order) = @_;
   $self->{decl_order} = $order;
