@@ -2,8 +2,6 @@
 
 package HTML::WebMake::FormatConvert;
 
-use Pod::Html;
-
 ###########################################################################
 # Define the converters we support here.
 # The method used is as follows:
@@ -29,10 +27,19 @@ sub set_converters {
   			'Text::EtText::EtText2HTML', \&et_to_html);
 
   $self->add_converter ('text/pod', 'text/html',
-  			'Pod::Html', \&pod_to_html);
+  			undef, \&pod_to_html);
+
+  $self->add_converter ('text/mshtml', 'text/html',
+  			undef, \&mshtml_to_html);
 
   $self->add_converter ('text/html', 'text/plain',
   			undef, \&html_to_plain);
+
+  $self->add_converter ('text/xml', 'text/html',
+  			undef, \&xml_to_html);
+
+  $self->add_converter ('text/html', 'text/xml',
+  			undef, \&html_to_xml);
 }
 
 ###########################################################################
@@ -169,11 +176,11 @@ sub et_to_html {
   my ($self, $contobj, $txt) = @_;
 
   if (!defined $self->{ettext}) {
-    eval '
-      use Text::EtText::EtText2HTML;
+    eval {
+      require Text::EtText::EtText2HTML;
       $self->{ettext} = new Text::EtText::EtText2HTML;
-    1;' or
-    die "FormatConvert: cannot create Text::EtText::EtText2HTML object: $!";
+      1;
+    } or die "FormatConvert: cannot create Text::EtText::EtText2HTML object: $!";
 
     $self->{ettext}->{glossary} = $self->{main}->getglossary();
     $self->{ettext}->set_option ('EtTextHrefsRelativeToTop', '1');
@@ -205,14 +212,21 @@ sub pod_to_html {
   open (POD_OUT, "+>$tmpout") or die "Cannot write to $tmpout";
   my $start = tell(POD_OUT);
 
-  pod2html ('--infile='.$tmpin, '--outfile='.$tmpout, '--title=x', @args);
+  eval {
+    require Pod::Html;
+    Pod::Html::pod2html ('--infile='.$tmpin,
+    			'--outfile='.$tmpout, '--title=x', @args);
+  };
+
+  if ($@) { die "Pod::Html::pod2html failed: $@"; }
 
   seek (POD_OUT, $start, 0);
   $_ = join ('', <POD_OUT>);
   close POD_OUT;
 
   unlink ($tmpin, $tmpout);
-  unlink ("pod2htmd.x~~");	# more pod spoor
+  unlink ("pod2htmi.x~~");	# more pod spoor
+  unlink ("pod2htmd.x~~");
   unlink ("pod2html.x~~");
 
   # And now, some POD cleaning; the POD HTML isn't great unfortunately.
@@ -249,4 +263,46 @@ sub html_to_plain {
 
 # -------------------------------------------------------------------------
 
+sub html_to_xml {
+  my ($self, $contobj, $txt) = @_;
+  return $txt;		# super simple
+}
+
+# -------------------------------------------------------------------------
+
+sub xml_to_html {
+  my ($self, $contobj, $txt) = @_;
+  return $txt;		# super simple
+}
+
+# -------------------------------------------------------------------------
+
+sub mshtml_to_html {
+  my ($self, $contobj, $txt) = @_;
+
+  eval {
+    require HTML::WebMake::MSHTMLtoHTML;
+    $txt = HTML::WebMake::MSHTMLtoHTML::convert ($self, $contobj, $txt);
+  };
+  
+  if ($@) {
+    die "failed to convert mshtml_to_html: $@";
+  }
+
+  return $txt;
+}
+
+# -------------------------------------------------------------------------
+
+sub format_is_binary {				# static
+  my ($fmt) = @_;
+
+  if (!defined $fmt) { carp "undef arg in format_is_binary"; }
+  return 0 if ($fmt =~ m,^text/,);
+  return 0 if ($fmt =~ m,^application/xml,);
+  return 1;		# default
+}
+
+# -------------------------------------------------------------------------
+
 1;
