@@ -9,22 +9,23 @@ use HTML::Parser;
 use HTML::WebMake::Main;
 
 use vars	qw{
-  	@ISA
+  	@ISA $VERSION
 	@ALLFEATURES $INLINE_TAGS $KEEP_FORMAT_TAGS
 	$EMPTY_ELEMENT_TAGS $BOOL_ATTR_VALUE
 };
 
 @ISA = qw(HTML::Parser);
 
+$VERSION = 'bogus';     # avoid CPAN picking up html::parser ref
 
 @ALLFEATURES =		qw{
     pack nocomments addimgsizes addxmlslashes fixcolors cleanattrs
-    indent fixhrefs
+    indent fixhrefs 
 };
 
 $KEEP_FORMAT_TAGS =	qr{(?:xmp|listing|pre|plaintext)};
 
-$INLINE_TAGS =		qr{(?:a|b|i|em|q|strong|h\d|code|abbr|acronym|address|big|cite|del|ins|s|small|strike|sub|sup|u|samp|kbd|var|img|span)};
+$INLINE_TAGS =		qr{(?:a|b|i|em|q|strong|h\d|code|abbr|acronym|address|big|cite|del|ins|s|small|strike|sub|sup|u|samp|kbd|var|img|span|textarea)};
 
 $EMPTY_ELEMENT_TAGS =	qr{(?:area|base|basefont|bgsound|br|col|embed|frame|hr|img|input|isindex|keygen|link|meta|param|spacer|wbr)};
 
@@ -44,10 +45,14 @@ sub new {
   # this parameter is not supported in earlier versions
   if ($HTML::Parser::VERSION >= 3.00) {
     my $val = '==BOOL_TRUE==';
-    eval '
+    eval {
       $self->boolean_attribute_value ($val);
       $BOOL_ATTR_VALUE = $val;
-    ';
+    };
+
+    # make sure we get the tokenpos: this is a hack so we can tell
+    # if we're parsing real XML instead of HTML
+    $self->handler ( end => "end", "self, tagname, text, tokenpos");
   }
 
   bless ($self, $class);
@@ -99,6 +104,13 @@ sub clean {
   $self->{last_was_noninline_close_tag} = 0;
   $self->{last_text_was_whitespace} = 0;
 
+  if ($fname !~ /\..?htm?/i) {
+    # use XML mode if possible
+    if ($self->can ("xml_mode")) { $self->xml_mode (1); }
+    # preserve case: important for RDF
+    if ($self->can ("case_sensitive")) { $self->case_sensitive (1); }
+  }
+
   $self->parse ($$txt); $self->eof();
 
   if ($self->{indent_level} > 0) {
@@ -137,7 +149,7 @@ sub start {
   if (!$is_inline_tag && !$self->{in_pre}) {
     $self->add_text ("\n");
     if (!$self->{in_pre}) {
-      if ($tagname !~ /^${EMPTY_ELEMENT_TAGS}$/) {
+      if ($tagname !~ /^${EMPTY_ELEMENT_TAGS}$/ && $origtext !~ /\/>$/) {
 	$self->open_indent();
       } else {
 	$self->add_current_indent();
@@ -187,8 +199,9 @@ sub clean_attrs_at_start {
   }
 
   my $tagend = '';
-  if ($self->{addxmlslashes} && $attrs !~ /\/\s*$/ &&
+  if ($attrs !~ /\/\s*$/ && (($self->{addxmlslashes} &&
 	$tagname =~ /^${EMPTY_ELEMENT_TAGS}$/)
+      || $origtext =~ /\s\/>$/))
   {
     $tagend = " />";		# HTML-4, XHTML, XML style
   } else {
@@ -213,17 +226,23 @@ sub clean_attrs_at_start {
 # --------------------------------------------------------------------------
 
 sub end {
-  my($self, $tagname, $origtext) = @_;
+  my($self, $tagname, $origtext, $tokenpos) = @_;
+
+  # XML <foo /> tags
+  if (!defined ($tokenpos)) { return; }
   
   my $exiting_pre = ($tagname =~ /^${KEEP_FORMAT_TAGS}$/);
   if ($exiting_pre) { $self->{in_pre}--; }
+  $self->{last_text_was_whitespace} = 0;
 
-  if ($tagname !~ /^${INLINE_TAGS}$/ && !$exiting_pre) {
+  if ($self->{in_pre}) {
+    $self->add_text ($origtext);
+
+  } elsif ($tagname !~ /^${INLINE_TAGS}$/ && !$exiting_pre) {
     if (!$self->{last_was_noninline_close_tag}) {
       $self->add_text ("\n");
     }
     $self->close_indent();
-
     $self->add_text ("</$tagname>\n");
     $self->{last_was_noninline_close_tag} = 1;
 
@@ -231,7 +250,6 @@ sub end {
     $self->add_text ("</$tagname>");
     $self->{last_was_noninline_close_tag} = 0;
   }
-  $self->{last_text_was_whitespace} = 0;
 }
 
 # --------------------------------------------------------------------------
