@@ -69,6 +69,7 @@ use HTML::WebMake::Metadata;
 use HTML::WebMake::PerlCode;
 use HTML::WebMake::FormatConvert;
 use HTML::WebMake::DataSource;
+use HTML::WebMake::CGIGlue;
 use HTML::WebMake::SiteMap;
 use HTML::WebMake::UserTags;
 use HTML::WebMake::WMLinkGlossary;
@@ -95,7 +96,7 @@ $SUBST_META		= '(!M)';
 
 ###########################################################################
 
-=item $f = new HTML::WebMake::Main
+=item $f = new HTML::WebMake::Main({ ... })
 
 Constructs a new C<HTML::WebMake::Main> object.  You may pass the following
 attribute-value pairs to the constructor.
@@ -150,6 +151,11 @@ sub new {
   $class = ref($class) || $class;
 
   my $self = shift;
+  if (!defined $self) {
+    $self = { };
+  }
+  bless ($self, $class);
+
   $self->{paranoid}		||= 0;
   $self->{debug}		||= 0;
   if (!defined $self->{verbose}) { $self->{verbose} = 1; }
@@ -178,6 +184,8 @@ sub new {
   $self->{locations}	= { };
   $self->{location_order} = [ ];
 
+  $self->{actions}	= { };
+
   # increase the size of these hashes in anticipation of big filesets
   keys %{$self->{outs}} = 300;
   keys %{$self->{locations}} = 300;
@@ -202,12 +210,15 @@ sub new {
   $self->{util}		= new HTML::WebMake::Util();
   $self->{have_image_size_module} = undef;
 
-  $self->{subst_stack} = [ ];
-  $self->{current_subst} = undef;
-  $self->{dep_datasources} = { };
+  $self->{subst_stack}		= [ ];
+  $self->{current_subst}	= undef;
+  $self->{dep_datasources}	= { };
   $self->{current_webmake_fname} = undef;
+  $self->{metaset}		= 'def';
 
-  bless ($self, $class);
+  $self->{xform_hooks}		= [ ];
+
+  $self->{lang}			= undef;
 
   $DEBUG = $self->{debug};
   $VERBOSE = $self->{verbose};
@@ -225,16 +236,21 @@ sub init_for_making {
   if ($^O !~ /(win|os2|mac)/i) {
     # which genius decided the mere sniff of getpwuid() should cause
     # a crash on win32? Cheers mate.
-    eval ' $self->{current_user} = getpwuid ($>); ';
+    eval {
+      $self->{current_user} = getpwuid ($>);
+    };
   } else {
     $self->{current_user} = "unknown";
   }
 
   $self->{format_conv} = new HTML::WebMake::FormatConvert ($self);
   $self->{metadata} = new HTML::WebMake::Metadata ($self);
+  $self->{cgiglue} = new HTML::WebMake::CGIGlue ($self);
 
   $self->{ignore_for_dependencies} =
   			new HTML::WebMake::File ($self, $SUBST_DEP_IGNORE);
+  $self->{always_rebuild_for_dependencies} =
+  			new HTML::WebMake::File ($self, $SUBST_EVAL);
   $self->{meta_ignore_for_dependencies} =
   			new HTML::WebMake::File ($self, $SUBST_META);
 
@@ -334,6 +350,14 @@ sub tmpdir {
   $self->{seddedtmpdir} = $fname;
 }
 
+sub tmpfile {
+  my ($self, $type, $ext) = @_;
+
+  my $tmpdir = $self->tmpdir();
+  my $rand = int(rand(99999)) . '_' . $$;
+  return File::Spec->catdir ($tmpdir, "TMP_${type}_${rand}.$ext");
+}
+
 # -------------------------------------------------------------------------
 
 sub cachedir {
@@ -424,11 +448,12 @@ sub readfile {
   my $wmkf = new HTML::WebMake::WmkFile ($self, $fname);
 
   if (defined $opts && $opts->{parse_for_cgi}) {
-    $wmkf->{parse_for_cgi} = 1;
+    $wmkf->set_parse_for_cgi();
   }
 
   $wmkf->parse ($_);
   $self->{files}->{$fname} = $wmkf;
+  $self->{cgi_data} = $wmkf->{cgi};
 
   $self->add_source_files ($fname);
   1;
@@ -458,6 +483,21 @@ sub readstring {
 
 # -------------------------------------------------------------------------
 
+=item $str = $f->get_content ($name);
+
+Get the item of content named C<$name>.  Equivalent to a $ {content_reference},
+and equivalent to the same method in C<HTML::WebMake::PerlCodeLibrary>.
+
+=cut
+
+sub get_content {
+  my ($self, $key) = @_;
+  if (!defined $key) { croak ("get_content with undef key"); }
+  return $self->curly_or_meta_subst ($SUBST_EVAL, $key);
+}
+
+# -------------------------------------------------------------------------
+
 # Internal API, used by HTML::WebMake::CGI modules.  This parses the
 # .wmk file (quickly) and generates a list of the editable items therein.
 #
@@ -465,7 +505,7 @@ sub cgi_parse_file {
   my ($self, $fname, $opts) = @_;
 
   if ($self->readfile ($fname, { 'parse_for_cgi' => 1 })) {
-    return $self->{files}->{$fname}->{cgi};
+    return $self->{cgi_data};
   } else {
     return undef;
   }
@@ -520,7 +560,7 @@ sub add_out {
 
 sub set_metadata_content_item ($$$$$) {
   my ($self, $name, $file, $attrs, $text) = @_;
-
+  dbg2("Main->set_metadata_content_item : $name");
   if (!defined $self->{metadatas}->{$name}) {
     push (@{$self->{content_order}}, $name);
   }
@@ -566,7 +606,7 @@ sub set_unmapped_content ($$$) {
   dbg2 ("set \"$key\" (unmapped)");
 
   return new HTML::WebMake::NormalContent ($key,
-  	$self->{ignore_for_dependencies},
+  	$self->{always_rebuild_for_dependencies},
 	  {
 	    'format'		=> 'text/html',
 	    'map'		=> 'false',
@@ -652,13 +692,31 @@ sub add_metadata {
 
   $attrs->{up} = $from;
 
-  if ($setthisdotmetas) {
-    $self->set_metadata_content_item ($thiskey, $wmkf, $attrs, $val);
-    push (@{$self->{this_metas_added}}, $thiskey);
-  }
+  if (defined($attrs->{lang}) ){
+    dbg2 ("metadata lang  attr == \"$attrs->{lang}\"") ;
+    my $langkey = "::".lc($attrs->{lang});
+    dbg2 ("set metadata $key == \"".$fullkey.$langkey."\"");
 
-  $self->set_metadata_content_item ($fullkey, $wmkf, $attrs, $val);
-  $self->getcache()->put_metadata ($fullkey, $val);
+	if ($setthisdotmetas) {
+	  $self->set_metadata_content_item ($thiskey.$langkey, $wmkf, $attrs, $val);
+	  push (@{$self->{this_metas_added}}, $thiskey.$langkey);
+	}
+
+    $self->set_metadata_content_item ($fullkey.$langkey, $wmkf, $attrs, $val);
+    $self->getcache()->put_metadata ($fullkey.$langkey, $val);
+
+    #return if ($self->{lang} !~ $attrs->{lang});
+
+  } else { 
+
+	if ($setthisdotmetas) {
+	  $self->set_metadata_content_item ($thiskey, $wmkf, $attrs, $val);
+	  push (@{$self->{this_metas_added}}, $thiskey);
+	}
+
+	$self->set_metadata_content_item ($fullkey, $wmkf, $attrs, $val);
+	$self->getcache()->put_metadata ($fullkey, $val);
+  }
 }
 
 # -------------------------------------------------------------------------
@@ -725,6 +783,10 @@ sub gc_contents {
 
 sub add_url {
   my ($self, $name, $location) = @_;
+
+  my $lang = $self->{current_subst}->{lang};
+  $name  = $name."::".$lang if defined $lang;
+
   dbg2 ("adding URL \"$name\" = $location");
   if (!defined $self->{locations}->{$name}) {
     push (@{$self->{location_order}}, $name);
@@ -786,7 +848,6 @@ sub add_navlinks {
   $self->curly_subst ($HTML::WebMake::Main::SUBST_EVAL, $map);
   $self->getmapper()->{set_navlinks} = 0;
 
-
   $attrs->{nav_up} = $attrs->{up};
   $attrs->{nav_next} = $attrs->{next};
   $attrs->{nav_prev} = $attrs->{prev};
@@ -826,6 +887,13 @@ sub add_breadcrumbs {
 
 # -------------------------------------------------------------------------
 
+sub add_action {
+  my ($self, $event, $code) = @_;
+  $self->{actions}->{$event} = $code;
+}
+
+# -------------------------------------------------------------------------
+
 sub subst {
   my ($self, $from, $str, $evaluatingtags) = @_;
 
@@ -840,6 +908,7 @@ sub subst {
     $self->infinite_subst_loop_error ($from, $$str);
     return "";
   }
+  my $val;
 
   {
     $current_subst->{level}++;
@@ -856,20 +925,54 @@ sub subst {
     if ($$str =~ /\$\{IMGSIZE\}/is) {
       # magic tag: <img src=foo.gif ${IMGSIZE}>
       $$str =~ s/<img\s+([^>]*?)\s*\$\{IMGSIZE\}\s*([^>]*?)\s*>/
-		  $self->add_image_size ($from, $1, $2);
-	  /gies;
+      		$self->add_image_size ($from, $1, $2);
+      /gies;
     }
 
-    # references to content chunks: ${content}
-    $$str =~ s/\$\{([^\<\{\}]+)\}/ $self->_curly_subst ($from, $1, 1) /ges;
-    #}
+    # rather than using s///ge to expand the references, it's a little
+    # faster to use a lexical scanner.  Not much, but a little...
 
-    # references to out URLs: $(foo)
-    $$str =~ s/\$\(([^\<\(\)]+)\)/ $self->_round_subst ($from, $1); /ges;
+#    # references to content chunks: ${content}
+#    $$str =~ s/\$\{([^\<\{\}]+)\}/ $self->_curly_subst ($from, $1, 1) /ges;
+#    # }
+#
+#    # references to out URLs: $(foo)
+#    $$str =~ s/\$\(([^\<\(\)]+)\)/ $self->_round_subst ($from, $1) /ges;
+#
+#    # references to metadata: $[this.foo] used within the chunk they're
+#    # defined in.
+#    $$str =~ s/\$\[this(\.[^\[\]]+)\]/ $self->_this_subst ($from, $1) /gies;
 
-    # references to metadata: $[this.foo] used within the chunk they're
-    # defined in.
-    $$str =~ s/\$\[this(\.[^\[\]]+)\]/ $self->_this_subst ($from, $1); /gies;
+    my @newstr = ();
+    LOOP:
+    {
+      # references to content chunks: ${content}
+      if ($$str =~ /\G\$\{([^\<\{\}]+)\}/gcs) {
+	push (@newstr, $self->_curly_subst ($from, $1, 1));
+	redo LOOP;
+      }
+
+      # references to out URLs: $(foo)
+      if ($$str =~ /\G\$\(([^\<\(\)]+)\)/gcs) {
+	push (@newstr, $self->_round_subst ($from, $1));
+	redo LOOP;
+      }
+
+      # references to metadata: $[this.foo] used within the chunk they're
+      # defined in.
+      if ($$str =~ /\G\$\[this(\.[^\[\]]+)\]/gcis) {
+        push (@newstr, $self->_this_subst ($from, $1));
+        redo LOOP;
+      }
+
+      # anything else.
+      if ($$str =~ /\G(\$?[^\$]+)/gcs) {
+        push (@newstr, $1);
+        redo LOOP;
+      }
+    };
+
+    $$str = join ('', @newstr);
 
 done_substs:
 
@@ -893,15 +996,16 @@ sub subst_deferred_refs {
     }
 
     # deferred refs to content chunks: $[content]
+    # (note that using /ges is *faster* than a lex scanner)
     $$str =~ s/\$\[([^\[\]]+)\]/ $self->_curly_subst ($from, $1, 0); /ges;
-
+    
     # do a subst in case the deferred ref contained normal refs
     $self->subst ($from, $str);
 
   } while ($$str =~ /\$\{.*?\}/ ||
 	  $$str =~ /\$\[.*?\]/ ||
 	  $$str =~ /\$\(.*?\)/ ||
-	  $$str =~ /<\{.*?\>\}/);
+	  $$str =~ /<\{.*?\}\>/);
 }
 
 # -------------------------------------------------------------------------
@@ -918,7 +1022,7 @@ sub infinite_subst_loop_error {
     $err =~ s/^.{6,}?(.{0,16}\$\{.*?\}.{0,16}).*?$/\[...\]$1\[...\]/gs;
     $err =~ s/^.{6,}?(.{0,16}\$\[.*?\].{0,16}).*?$/\[...\]$1\[...\]/gs;
     $err =~ s/^.{6,}?(.{0,16}\$\(.*?\).{0,16}).*?$/\[...\]$1\[...\]/gs;
-    $err =~ s/^.{6,}?(.{0,16}\<\{.*?\>\}.{0,16}).*?$/\[...\]$1\[...\]/gs;
+    $err =~ s/^.{6,}?(.{0,16}\<\{.*?\}\>.{0,16}).*?$/\[...\]$1\[...\]/gs;
     # $err =~ s/^.{6,}?(.{0,16}(?:\$[\{\(\[]|\<\{).{0,16}).*?$/\[...\]$1\[...\]/gs;
   }
 
@@ -940,16 +1044,17 @@ sub infinite_subst_loop_error {
 
 sub fileless_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open(undef, undef, undef, "text/html", 0);	#{
+  $self->_subst_open(undef, undef, undef, "text/html", 0, undef);	#{
   $self->subst($from, \$txt);
   $self->strip_metadata ($from, \$txt);
+  $self->subst_deferred_refs ($from, \$txt);
   $self->_subst_close();				#}
   $txt;
 }
 
 sub curly_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open (undef, undef, undef, "text/html", undef);	#{
+  $self->_subst_open (undef, undef, undef, "text/html", undef, undef);	#{
   $txt = $self->_curly_subst ($from, $txt, 1);
   # then do a normal subst to handle <{set}>, metadata, etc.
   $self->subst ($from, \$txt);
@@ -960,7 +1065,7 @@ sub curly_subst {
 
 sub curly_meta_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open (undef, undef, undef, "text/html", undef);	#{
+  $self->_subst_open (undef, undef, undef, "text/html", undef, undef);	#{
   $txt = $self->_curly_subst ($from, $txt, 0);
   $self->_subst_close();				#}
   $txt;
@@ -968,7 +1073,7 @@ sub curly_meta_subst {
 
 sub curly_or_meta_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open (undef, undef, undef, "text/html", undef);	#{
+  $self->_subst_open (undef, undef, undef, "text/html", undef, undef);	#{
   $txt = $self->_curly_subst ($from, $txt, 2);
   # then do a normal subst to handle <{set}>, metadata, etc.
   $self->subst ($from, \$txt);
@@ -979,7 +1084,7 @@ sub curly_or_meta_subst {
 
 sub quiet_curly_meta_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open(undef, undef, undef, "text/html", undef);	#{
+  $self->_subst_open(undef, undef, undef, "text/html", undef, undef);	#{
   $self->{current_subst}->{quiet} = 1;
   $txt = $self->_curly_subst($from, $txt, 0);
   $self->{current_subst}->{quiet} = 0;
@@ -989,7 +1094,7 @@ sub quiet_curly_meta_subst {
 
 sub round_subst {
   my ($self, $from, $txt) = @_;
-  $self->_subst_open(undef, undef, undef, "text/html", undef);	#{
+  $self->_subst_open(undef, undef, undef, "text/html", undef, undef);	#{
   $txt = $self->_round_subst($from, $txt);
   $self->_subst_close();				#}
   $txt;
@@ -997,7 +1102,7 @@ sub round_subst {
 
 sub _this_subst {
   my ($self, $from, $origkey) = @_;
-
+  dbg2("Main->_this_subst: $origkey");
   # trim off the default value from our working copy of the key
   my $key = $origkey;
   $key =~ s/\?([^\?]+)$//;
@@ -1016,8 +1121,9 @@ sub _this_subst {
 
 # -------------------------------------------------------------------------
 
-sub _subst_open {
-  my ($self, $filename, $outname, $dotdots, $fmt, $useurls) = @_;
+sub _subst_open ($$$$$$$$) {
+  my ($self, $filename, $outname, $dotdots, $fmt, $useurls, $metaset,
+  	  $lang) = @_;
 
   my $current_subst = $self->{current_subst};
   if (defined $current_subst) {
@@ -1037,15 +1143,27 @@ sub _subst_open {
     if (!defined $useurls) {
       $useurls = $current_subst->{useurls};
     }
+    if (!defined $metaset) {
+      $metaset = $current_subst->{metaset};
+    }
+    if (!defined $lang) {
+      $lang = $current_subst->{lang};
+    }
   }
 
   # if (!defined $dotdots) { $dotdots = ""; }
-  if (!defined $filename) { $filename = $SUBST_EVAL; }
-  if (!defined $outname) { $outname = $SUBST_EVAL; }
-  if (!defined $useurls) { $useurls = 1; }
+  if (!defined $filename)  { $filename = $SUBST_EVAL; }
+  if (!defined $outname)   { $outname = $SUBST_EVAL; }
+  if (!defined $useurls)   { $useurls = 1; }
+  if (!defined $metaset)   { $metaset = $self->{metaset}; }
+  if (!defined $lang)      { $lang = $self->{lang}; }
 
-  $self->{current_subst} = new HTML::WebMake::SubstCtx
-	    ($self, $filename, $outname, $dotdots, $fmt, $useurls);
+  dbg2("Main->_subst_open: lang $lang") if defined $lang;
+  dbg2("Main->_subst_open: lang NOLANG") if !defined $lang;
+  dbg2("Main->_subst_open: metaset $metaset") if defined $metaset;
+  dbg2("Main->_subst_open: metaset NOSET") if !defined $metaset;
+  $self->{current_subst} = new HTML::WebMake::SubstCtx ($self,
+  		$filename, $outname, $dotdots, $fmt, $useurls, $metaset, $lang);
 }
 
 sub _subst_close {
@@ -1061,7 +1179,7 @@ sub _curly_subst {
   # if (!defined $from) { croak "No from defined in subst"; }
   # if (!defined $key) { croak "No key defined in subst"; }
 
-  # warn "JMD CURLY $key";
+  dbg2 ("_curly_subst:$from:$key:$contents_only");
 
   my $str;
   my $current_subst = $self->{current_subst};
@@ -1082,15 +1200,29 @@ sub _curly_subst {
   }
 
   my $cont;
+  
+  my $lang = $self->{current_subst}->{lang};
+  my $lang_key  = $key."::".$lang if defined $lang;
 
   if ($contents_only) {			# expanding a ${foo} ref
     $cont = $self->{contents}->{$key};
+    if (!defined $cont && defined($lang_key)) {
+      $cont = $self->{metadatas}->{$lang_key};
+  	  dbg2("Main->_curly_subst(0): cont  $lang_key") if (defined($cont));
+    }
     if (!defined $cont) {
       $cont = $self->{metadatas}->{$key};
+  	  dbg2("Main->_curly_subst(0): metas  $key") if (defined($cont));
     }
 
   } else {				# expanding a $[foo] ref
-    $cont = $self->{metadatas}->{$key};
+    $cont = $self->{metadatas}->{$lang_key} if defined($lang_key);
+    dbg2("Main->_curly_subst(1): metas1  $lang_key") if (defined($cont));
+
+    if (!defined $cont){
+	$cont = $self->{metadatas}->{$key};
+	    dbg2("Main->_curly_subst(1): metas2 $key") if (defined($cont));
+    } 
 
     # it's also possible to refer to content items using the metadata reference
     # type $[..], as in fact that reference type simply means a reference whose
@@ -1100,6 +1232,7 @@ sub _curly_subst {
     # metadata hash.
     if (!defined $cont) {
       $cont = $self->{contents}->{$key};
+  	  dbg2("Main->_curly_subst(1): contents $key") if (defined($cont));
     }
   }
 
@@ -1145,6 +1278,7 @@ sub _curly_subst {
   }
 
 ret:
+  dbg2 ("_curly_subst:end:$str:");
   $str;
 }
 
@@ -1158,10 +1292,15 @@ sub _round_subst {
   if (!defined $from) { croak "No from defined in subst"; }
   if ($self->{current_subst}->{inf_loop}) { return ""; }
 
+  my $lang = $self->{current_subst}->{lang};
+  my $lang_key ;
+
   my $defval = undef;
   if ($key =~ s/\?([^\?]+)$//) { $defval = $1; }
 
   my $str;
+  my $lang_str;
+
   if ($key eq 'TOP/') { $str = ''; }
   
   if (!defined $str) {
@@ -1172,16 +1311,34 @@ sub _round_subst {
       $self->subst_deferred_refs ($from, \$key);
     }
 
-    $str = $self->{locations}->{$key};
+    if (defined $lang) {
+      $lang_key  = $key."::".$lang;
+      $lang_str = $self->{locations}->{$lang_key};
+    } 
+    dbg2("Main->_round_subst \"$lang_key\" -> \"$lang_str\"") if defined $lang_str;
+    if (!defined $lang_str) {
+      $str = $self->{locations}->{$key};
+      dbg2("Main->_round_subst \"$key\" -> \"$str\"") if defined $str;
+    }
   }
 
+   $str = $lang_str if defined $lang_str;
+   
   if ((!defined $str || $str eq '') && $key ne 'TOP/') {
     if (defined $defval) { return $defval; }
-    vrb ("no value defined for output URL \$($key) in \"$from\".");
+    if (defined $lang_key) {
+      vrb ("no value defined for output URL \$($key) or \$($lang_key) in \"$from\".");
+    } else {
+      vrb ("no value defined for output URL \$($key) in \"$from\".");
+    }
     return "";
   }
 
-  $self->add_url_dependency ($key);
+  if (defined $lang_str) {
+    $self->add_url_dependency ($lang_key);
+  } else {
+    $self->add_url_dependency ($key);
+  }
 
   # make it a valid relative URL
   if ($str !~ /^\// && $str !~ /^[-_a-zA-Z0-9]:/) {
@@ -1217,6 +1374,8 @@ sub subst_metadata {
   # metadata must have the format "blah.type"
   return "" unless ($key =~ /^(.*)\.([^\.]+?)$/);
   my ($base, $subkey) = ($1, $2);
+  $subkey = lc $subkey;		# metas should be lowercase
+  $key = $base. '.' .$subkey;
 
   if ($from eq $base) { goto failed_to_find; }
 
@@ -1258,6 +1417,20 @@ sub subst_metadata {
     if (defined $meta) {
       return $self->metadata_to_content ($from, $key, $meta, $cont);
     }
+
+  } else {
+    # see if we can load a $[this.foo] by searching all the contents
+    # we've used so far.
+    foreach my $cont ($self->get_contents_used_in_this_output()) {
+      if (defined $cont->{implicit_metas}->{$subkey}) {
+	$cont->load_metadata ($base, $key);
+	$self->add_content_dependency ($cont);
+	my $meta = $self->getcache()->get_metadata ($key);
+	if (defined $meta) {
+	  return $self->metadata_to_content ($from, $key, $meta, $cont);
+	}
+      }
+    }
   }
 
 failed_to_find:
@@ -1309,7 +1482,9 @@ sub get_deferred_builtin_content {
   my ($self, $from, $key) = @_;
 
   if ($key eq "WebMake.Time") {
-    return strftime "%a %b %e %H:%M:%S %Y", localtime();
+    return q{<!--WMDepIgnore-->}.
+            strftime ("%a %b %e %H:%M:%S %Y", localtime()).
+            q{<!--/WMDepIgnore-->};
   }
   if ($key eq "WebMake.OutFile") {
     return $self->{current_subst}->{filename};
@@ -1326,6 +1501,9 @@ sub get_deferred_builtin_content {
   if ($key eq "WebMake.GeneratedFiles") {
     return join (' ', $self->generated_file_list());
   }
+  if ($key eq 'WebMake.ChangedFiles') {
+    return join (' ', @{$self->{made_files}});
+  }
   undef;
 }
 
@@ -1396,21 +1574,17 @@ sub eval_code_at_ref {
   }
 
   $self->{last_perl_code_text} = undef;
-  $$str =~ s/\<\{set\s*name\s*=\s*(.+?)\s+value\s*=\s*(.+?)\s*\}\>/
+  while (($$str =~ s/\<\{set\s*name\s*=\s*(.+?)\s+value\s*=\s*(.+?)\s*\}\>//is)
+  	|| ($$str =~ s/\<\{set\s*(.+?)\s*=\s*\"(.+?)\"\s*\}\>//is)
+  	|| ($$str =~ s/\<\{set\s*(.+?)\s*=\s*(.+?)\s*\}\>//is))
+  {
     $self->_eval_set ($from, $1, $2);
-  /gies;
-
-  $$str =~ s/\<\{set\s*(.+?)\s*=\s*\"(.+?)\"\s*\}\>/
-    $self->_eval_set ($from, $1, $2);
-  /gies;
-
-  $$str =~ s/\<\{set\s*(.+?)\s*=\s*(.+?)\s*\}\>/
-    $self->_eval_set ($from, $1, $2);
-  /gies;
+  }
 
   $$str =~ s/\<\{(perlout|perl)\s*(.+?)\s*\}\>/
     $self->_p_interpret($1, $2, '');
   /gies;
+
   $self->{last_perl_code_text};
 }
 
@@ -1474,11 +1648,14 @@ sub add_image_size {
   $attrtext =~ s/\s*\/\s*$//g;		# trim /> tag ending
 
   if (!defined $self->{have_image_size_module}) {
-    if (eval 'require Image::Size;') {
-      $self->{have_image_size_module} = 1;
-    } else {
+    eval {
+      require Image::Size;
+    };
+    if ($@) {
       vrb ("\${IMGSIZE} tag: cannot load Image::Size module, not supported");
       $self->{have_image_size_module} = 0;
+    } else {
+      $self->{have_image_size_module} = 1;
     }
   }
 
@@ -1525,11 +1702,11 @@ sub add_image_size {
   }
 
   $sizestr = '';
-  if (!eval '
-    use Image::Size qw(html_imgsize);
-    $sizestr = html_imgsize($realfname);
-    1;')
-  {
+  eval {
+    require Image::Size;
+    $sizestr = Image::Size::html_imgsize($realfname);
+  };
+  if ($@) {
     warn "\${IMGSIZE}: Image::Size failed: $! in \"$from\"\n";
     goto failed;
   }
@@ -1590,17 +1767,23 @@ sub canon_path {
   # and now trim off useless dir navigation like "foo/bar/../../baz"
   # down to "baz".
 
-  # first, deal with "foo/bar/../../[whatever]"
-  1 while $fname =~ s,^[^/][^\./]*?/+\.\./,,g;
+  while (1) {
+    my $got1 = 0;
 
-  # then "[whatever]/foo/bar/../../[whatever]"
-  1 while $fname =~ s,/[^/][^\./]*?/+\.\./,/,g;
+    # first, deal with "foo/bar/../../[whatever]"
+    $fname =~ s,^[^/][^\./]*?/+\.\./,,g and $got1++;
 
-  # then "[whatever]/foo/bar/../.."
-  1 while $fname =~ s,/[^/][^\./]*?/+\.\.$,,g;
+    # then "[whatever]/foo/bar/../../[whatever]"
+    $fname =~ s,/+[^/][^\./]*?/+\.\./,/,g and $got1++;
 
-  # and finally tidy up bonus slashes
-  $fname =~ s,//+,/,gs;
+    # then "[whatever]/foo/bar/../.."
+    $fname =~ s,/+[^/][^\./]*?/+\.\.$,,g and $got1++;
+
+    # and finally tidy up bonus slashes. doesn't count for $got1
+    $fname =~ s,//+,/,gs;
+
+    last unless $got1;
+  }
 
   return $fname;
 }
@@ -1665,6 +1848,11 @@ sub add_content_dependency {
 
   my $fname = $cont->get_filename();
 
+  # add it to the list of contents that could have metadata, if it can.
+  if (!$cont->is_generated_content()) {
+    $self->{contents_loaded}->{$cont->{name}} = $cont;
+  }
+
   if ($fname eq $SUBST_EVAL) {
     if ($self->{risky_fast_rebuild}) {
       dbg2 ("dependency: ". $cont->{name}.": [perl code, ignored]");
@@ -1706,6 +1894,12 @@ sub clear_dependencies {
   my ($self, $url) = @_;
 
   $self->{cont_dependencies} = { };
+  $self->{contents_loaded} = { };
+}
+
+sub get_contents_used_in_this_output {
+  my ($self) = @_;
+  return values (%{$self->{contents_loaded}});
 }
 
 # -------------------------------------------------------------------------
@@ -1755,6 +1949,7 @@ sub make {
 
   $self->{renames_required} = [ ];
   $self->{content_deps_required} = [ ];
+  $self->{made_files} = [ ];
 
   if ($#fnames < 0) {
     @fnames = @{$self->{out_order}};
@@ -1855,8 +2050,7 @@ sub make_file ($$) {
     croak ("no format defined for $fname");
   }
 
-  my $dotdots = '';
-  ($dotdots .= '../') while ($fname =~ m,[/\\],g);
+  my $dotdots = $self->make_dotdots_from_filename ($fname);
 
   my $useurls = 1;
   if (!$out->use_for_content_urls()) { $useurls = 0; }
@@ -1871,10 +2065,15 @@ sub make_file ($$) {
   }
   $self->{this_metas_added} = [ ];
 
-  $self->_subst_open($fname, $out->{name}, $dotdots, $fmt, $useurls);	#{
+  $self->_subst_open($fname, $out->{name}, $dotdots,
+  	$fmt, $useurls, $self->{metaset}, $out->{lang});		#{
+
   my $txt = $out->get_text();
-  $self->strip_metadata ($fname, \$txt);
-  $self->subst_deferred_refs ($fname, \$txt);
+
+  if (!HTML::WebMake::FormatConvert::format_is_binary ($fmt)) {
+    $self->strip_metadata ($fname, \$txt);
+    $self->subst_deferred_refs ($fname, \$txt);
+  }
 
   if ($txt =~ /{!!WMDEFER/) {
     $self->make_file_defer ($fname, $out, $outfname, $txt);
@@ -1889,11 +2088,37 @@ sub make_file ($$) {
 
 # -------------------------------------------------------------------------
 
+sub make_dotdots_from_filename ($$) {
+  my ($self, $fname) = @_;
+
+  my $curdir = getcwd();
+  if ($self->{base_dir} ne '') {
+    $curdir = $self->{base_dir};
+  }
+  $curdir = File::Spec->canonpath ($curdir);
+  my @dirstohere = split (m![/\\]+!, $curdir);
+
+  my $dotdots = '';
+  while ($fname =~ m!(.*?)[/\\]+!g) {
+    if ($1 eq '.') {
+      next;
+    } elsif ($1 eq '..') {
+      my $dir = pop (@dirstohere);
+      $dotdots = $dir.'/'.$dotdots;
+    } else {
+      push (@dirstohere, $1);
+      $dotdots = '../'.$dotdots;
+    }
+  }
+  $dotdots;
+}
+
+# -------------------------------------------------------------------------
+
 sub make_file_finish ($$$) {
   my ($self, $fname, $out, $outfname, $txt) = @_;
 
-  my $dotdots = '';
-  ($dotdots .= '../') while ($fname =~ m,[/\\],g);
+  my $dotdots = $self->make_dotdots_from_filename ($fname);
 
   # unescape escaped references to our entities.
   $txt =~ s/\&wmdollar;/\$/gis;
@@ -1923,15 +2148,26 @@ sub make_file_finish ($$$) {
   if (!$self->{making_to_string} && $self->{force_output} == 0 && -f $outfname)
   {
     my $curtxt;
-    if ((-s $outfname == length($txt))
-      	&& (open (IN, "<$outfname"))
+
+    if (
+        #(-s $outfname == length($txt)) &&      # will not work with DepIgnore
+        (open (IN, "<$outfname"))
 	&& ($curtxt = join ('', <IN>))
-	&& (close IN)
-	&& ($curtxt eq $txt)
-	)
+	&& (close IN))
     {
-      dbg ("not making (text has not changed): $outfname");
-      return;
+      # trim out WMDepIgnore-comment-wrapped sections.  This is used
+      # for e.g. the last modification date, which will always differ
+      # on a rebuild!
+      #
+      $curtxt =~ s{<!--WMDepIgnore-->.*?<!--/WMDepIgnore-->}{}gs;
+
+      my $txt2 = $txt;
+      $txt2 =~ s{<!--WMDepIgnore-->.*?<!--/WMDepIgnore-->}{}gs;
+
+      if ($curtxt eq $txt2) {
+        dbg ("not making (text has not changed): $outfname");
+        return;
+      }
     }
   }
 
@@ -1955,6 +2191,7 @@ sub make_file_finish ($$$) {
       $self->fail ("Cannot write: $newfname"); return;
     }
 
+    push (@{$self->{made_files}}, $outfname);
     push (@{$self->{renames_required}}, [ $newfname, $outfname ]);
     push (@{$self->{content_deps_required}}, [ $fname,
 				      $self->{cont_dependencies} ]);
@@ -1985,17 +2222,34 @@ sub finish_deferred_files {
   foreach my $fname (keys %{$self->{need_rewrite_for_deferred_urls}})
   {
     dbg ("fixing URLs in deferred out file: $fname");
+
     my $txt = $self->{need_rewrite_for_deferred_urls}->{$fname};
     my $ctx = $self->{need_rewrite_subst_context}->{$fname};
 
-    $self->_subst_open($ctx->{filename}, $ctx->{outname},
-    		$ctx->{dotdots}, $ctx->{format}, $ctx->{useurls});
-
+    $self->_subst_open($ctx->{filename}, $ctx->{outname}, $ctx->{dotdots},
+    		$ctx->{format}, $ctx->{useurls}, $ctx->{metaset},
+			$ctx->{lang});
     #{
+
     $txt =~ s/{!!WMDEFER_dotdots}/$ctx->{dotdots}/gs;
-    $txt =~ s/{!!WMDEFER_content_url:([^}]+)}/
-      $self->rewrite_a_deferred_url($1, $give_up_if_still_deferred);
-    /ges;
+    my @defernames = ($txt =~ m/{!!WMDEFER_content_url:([^}]+)}/gs);
+
+    my $url;
+    foreach my $cname (@defernames) {
+      my $obj = $self->get_content_obj ($cname);
+
+      if (!defined $obj || !defined ($url = $obj->get_url())) {
+	$self->fail ("unable to get URL for content item: \${$cname}");
+	$url = '';
+
+      } elsif ($url eq "{!!WMDEFER_content_url:$cname}") {
+	next unless ($give_up_if_still_deferred);
+	$self->fail ("unable to get URL for content item: \${$cname}");
+	$url = '';
+      }
+
+      $txt =~ s/{!!WMDEFER_content_url:$cname}/$url/g;
+    }
 
     $self->_subst_close();
 
@@ -2024,29 +2278,11 @@ sub finish_deferred_files {
     %{$self->{need_rewrite_for_deferred_urls}} = %new_deferred_list;
     return 1;
   } else {
+    $self->{need_rewrite_for_deferred_urls} = { };
     return 0;
   }
 }
 
-sub rewrite_a_deferred_url {
-  my ($self, $contname, $give_up_if_still_deferred) = @_;
-
-  my $obj = $self->get_content_obj ($contname);
-
-  my $url;
-  if (!defined $obj || !defined ($url = $obj->get_url())) {
-    $self->fail ("unable to get URL for content item: \${$contname}");
-    return '';
-  }
-
-  if ($give_up_if_still_deferred && $url =~ /^\{!!WMDEFER_content_url:/)
-  {
-    $self->fail ("unable to get URL for content item: \${$contname}");
-    return '';
-  }
-  return $url;
-}
-
 sub make_deferred_url {
   my ($self, $contname) = @_;
   return '{!!WMDEFER_content_url:'.$contname.'}';
@@ -2124,11 +2360,11 @@ sub clean_html {
   if ($features !~ /\S/) { return $$txtptr; }
 
   if (!defined $self->{htmlcleaner}) {
-    if (!eval '
-	use HTML::WebMake::HTMLCleaner;
-	$self->{htmlcleaner} = new HTML::WebMake::HTMLCleaner($self);
-      1;')
-    {
+    eval {
+      require HTML::WebMake::HTMLCleaner;
+      $self->{htmlcleaner} = new HTML::WebMake::HTMLCleaner($self);
+    };
+    if ($@) {
       warn "HTMLCleaner load failed: $@\n";
       warn "HTMLCleaner load failed -- not cleaning HTML.\n";
       $self->{htmlcleaner} = { 'loadfailed' => 1 };
@@ -2168,6 +2404,14 @@ could not be created, etc.).
 sub finish {
   my ($self) = @_;
 
+  if (defined $self->{actions}->{site_changed}) {
+    my $action = $self->{actions}->{site_changed};
+    if (scalar @{$self->{made_files}} > 0) {
+      # just run the code
+      $self->eval_code_at_parse (\$action);
+    }
+  }
+
   if (defined $self->{cache}) {
     $self->{cache}->untie();
   }
@@ -2176,6 +2420,21 @@ sub finish {
 
 # -------------------------------------------------------------------------
 
+sub get_content_xform_hooks {
+  my ($self) = @_;
+  @{$self->{xform_hooks}};
+}
+
+sub add_content_xform_hook {
+  my ($self, $cb, @args) = @_;
+  push (@{$self->{xform_hooks}}, {
+    cb => $cb,
+    args => \@args
+  });
+}
+
+# -------------------------------------------------------------------------
+
 sub quicktxt2html {
   my $txt = join ('',@_);
 
@@ -2222,8 +2481,21 @@ sub stacktrace {
   carp join ("\n", @_);
 }
 
+# -------------------------------------------------------------------------
+
+sub get_editcgi {
+  my $self = shift;
+
+  my $editcgi = $self->{main}->{locations}->{'WebMake.EditCGI'};
+  if (!defined $editcgi) {
+    die "editbuttons: <editcgi> required before <use> line.\n";
+  }
+}
+
 1;
 
+# -------------------------------------------------------------------------
+
 __END__
 
 =back
