@@ -11,7 +11,7 @@ use strict;
 use HTML::WebMake::Main;
 
 use vars	qw{
-  	@ISA %BUILTIN_TYPES $NUM $STR
+  	@ISA %BUILTIN_TYPES $NUM $STR 
 };
 
 # -------------------------------------------------------------------------
@@ -47,11 +47,15 @@ sub new ($$$$$) {
 
   my $self = {
     'main'		=> $main,
+    'metasets'		=> { },
     'metadefaults'	=> { },
     'attrdefaults'	=> { },
   };
   bless ($self, $class);
 
+  # setup the default set of metadata types
+  $self->{metasets}->{'def'} = \%BUILTIN_TYPES;
+
   $self;
 }
 
@@ -59,23 +63,49 @@ sub dbg { HTML::WebMake::Main::dbg (@_); }
 
 # -------------------------------------------------------------------------
 
-sub get_type {
+sub get_current_metaset ($)
+{
+  my ($self) = @_;
+  ## CHRIS: FIXME added ->{metaset}, don't know why it's not defined
+  ##        when {current_subst} is
+  if (defined $self->{main}->{current_subst}->{metaset}) {
+    return $self->{main}->{current_subst}->{metaset};
+  } elsif (defined $self->{main}->{metaset}) {
+    return $self->{main}->{metaset};
+  } else {
+    return 'def';
+  }
+}
+
+# -------------------------------------------------------------------------
+
+sub get_type ($$)
+{
   my ($self, $meta) = @_;
 
   croak "no meta defined in get_type" unless defined($meta);
 
-  my $info = $BUILTIN_TYPES{$meta};
-  if (!defined $info) { return undef; }
+  $meta =~ s/::.*$//;
+
+  my $set = $self->get_current_metaset();
+  my $info = $self->{metasets}->{$set}->{$meta};
+
+  # default to string type for unknown names of metadata
+  if (!defined $info) { return $STR; }
   
   return $info->{type};
 }
 
-sub get_default_value {
+sub get_default_value ($$)
+{
   my ($self, $meta) = @_;
 
   croak "no meta defined in get_default_value" unless defined($meta);
 
-  my $info = $BUILTIN_TYPES{$meta};
+  $meta =~ s/::.*$//;
+
+  my $set = $self->get_current_metaset();
+  my $info = $self->{metasets}->{$set}->{$meta};
   if (!defined $info) { return undef; }
   
   return $info->{default};
@@ -83,13 +113,18 @@ sub get_default_value {
 
 # -------------------------------------------------------------------------
 
-sub convert_to_type {
+sub convert_to_type ($$$)
+{
   my ($self, $meta, $val) = @_;
 
   croak "no meta defined in get_type" unless defined($meta);
   croak "no val defined in get_type" unless defined($val);
 
-  my $info = $BUILTIN_TYPES{$meta};
+  $meta =~ s/::.*$//;
+
+  dbg("Metadata->convert_to_type: meta $meta");
+  my $set = $self->get_current_metaset();
+  my $info = $self->{metasets}->{$set}->{$meta};
   if (!defined $info) { return undef; }
   
   if ($info->{type} == $NUM) {
@@ -101,7 +136,8 @@ sub convert_to_type {
 
 # -------------------------------------------------------------------------
 
-sub string_to_sort_sub {
+sub string_to_sort_sub ($$)
+{
   my ($self, $sortstring) = @_;
 
   my @substrs = ();
@@ -209,4 +245,45 @@ sub get_attrdefault {
 
 # -------------------------------------------------------------------------
 
+sub parse_metaset {
+  my ($self, $id, $text, $attrs) = @_;
+  local ($_);
+
+# <metaset id="dc">
+# DC.Title: type=string default="(Untitled)"
+# </metaset>
+
+  $self->{metasets}->{$id} = { };
+  foreach $_ (split (/\n/s, $text)) {
+    next if /^\s*$/;
+    if (!/^\s*(\S+):\s+(.*)\s*$/) {
+      warn "invalid <metaset> line: $_\n"; next;
+    }
+
+    my $meta = lc $1;
+    my $attrs = $self->{main}->{util}->parse_xml_tag_attributes
+    			("metaset", $2, "<metaset>", qw(type default));
+    my $item;
+    if ($attrs->{type} eq 'string') {
+      $item = {
+		'type' => $STR,
+		'default' => $attrs->{default},
+      };
+
+    } elsif ($attrs->{type} eq 'numeric') {
+      $item = {
+		'type' => $NUM,
+		'default' => $attrs->{default} + 0,
+      };
+
+    } else {
+      warn "invalid <metaset> line, bad type: $_\n"; next;
+    }
+
+    $self->{metasets}->{$id}->{$meta} = $item;
+  }
+}
+
+# -------------------------------------------------------------------------
+
 1;
