@@ -60,6 +60,11 @@ sub new {
     delete $self->{'asis'};	# in case it was set as an attr
   }
 
+  # binary files cannot be subst'ed
+  if (HTML::WebMake::FormatConvert::format_is_binary ($self->get_format())) {
+    $self->{keep_as_is} = 1;
+  }
+
   $attrval = $attrs->{'isroot'};
   $attrval ||= $metadata->get_attrdefault ('isroot');
   if (defined $attrval) {
@@ -135,6 +140,13 @@ sub is_generated_content {
 
 # -------------------------------------------------------------------------
 
+sub is_template_content {
+  my ($self) = @_;
+  return $self->{no_map};
+}
+
+# -------------------------------------------------------------------------
+
 sub expand {
   my ($self) = @_;
   return $self->{main}->curly_subst ($self->{name}, $self->{name});
@@ -150,24 +162,61 @@ sub expand_no_ref {
 sub get_metadata {
   my ($self, $key) = @_;
 
+
   if (!defined $self->{cached_metas}) {
     $self->{cached_metas} = { };
   }
 
-  my $val = $self->{cached_metas}->{$key};
   my $main = $self->{main};
+  my $l = $self->{main}->{current_subst}->{lang};
+  $l = "undef" if ! defined($l);
 
-  if (!defined $val) {
-    $val = $main->quiet_curly_meta_subst
-    		($HTML::WebMake::Main::SUBST_META, $self->{name}.".".$key);
-    if (!defined $val) {
-      $val ||= $main->{metadata}->get_default_value ($key);
-    }
+  my $lang_key  = $key."::".$l;
+  my $val = undef;
 
+  ###########
+  $val = $self->{cached_metas}->{$lang_key};
+  dbg2("NormalContent->get_metadata(1)") if (defined($val));
+  goto ret if defined($val);
+
+  ###########
+  #$val = $main->quiet_curly_meta_subst
+  #		  ($HTML::WebMake::Main::SUBST_META, $self->{name}.".".$lang_key);
+  #dbg2("NormalContent->get_metadata(2): $val") if (defined($val));
+  #goto ret_conv2 if defined($val);
+
+  ###########
+  $val = $self->{cached_metas}->{$key};
+  dbg2("NormalContent->get_metadata(3)") if (defined($val));
+  goto ret if defined($val);
+
+  ###########
+  $val = $self->{implicit_metas}->{$key};
+  dbg2("NormalContent->get_metadata(4)") if (defined($val));
+  goto ret if defined($val);
+
+  ###########
+  $val = $main->quiet_curly_meta_subst
+		  ($HTML::WebMake::Main::SUBST_META, $self->{name}.".".$key);
+  dbg2("NormalContent->get_metadata(4)") if (defined($val));
+  goto ret_conv if defined($val);
+
+
+  ###########
+  $val ||= $main->{metadata}->get_default_value ($key);
+  dbg2("NormalContent->get_metadata(5)");
+
+ret_conv:
     $val = $main->{metadata}->convert_to_type ($key, $val);
     $self->{cached_metas}->{$key} = $val;
-  }
+	goto ret;
 
+ret_conv2:
+    $val = $main->{metadata}->convert_to_type ($key, $val);
+    $self->{cached_metas}->{$lang_key} = $val;
+
+ret:
+  #dbg2 ("NormalContent->get_metadata: \"$key\", lang=\"$l\", val=\"$val\"");
   return $val;
 }
 
@@ -180,15 +229,23 @@ sub create_extra_metas_if_needed {
   }
 }
 
+sub create_implicit_metas_if_needed {
+  my ($self) = @_;
+  if (!defined $self->{implicit_metas}) {
+    $self->{implicit_metas} = { };
+  }
+}
+
 # -------------------------------------------------------------------------
 
 sub load_metadata {
   my ($self, $name, $key) = @_;
 
-  # a different method is used to load metadata from the current content
-  # item, so this should not happen:
-  if ($key =~ /^this\./i) {
-    warn "oops! wasn't expecting a this. metaref in load_metadata: $key";
+  my $subkey;
+  if ($key =~ /\.([^\.]+?)$/) {
+    $subkey = lc $1;         # metas should be lowercase
+  } else {
+    $subkey = lc $key;
   }
 
   # unmapped content can't have metadata
@@ -211,6 +268,14 @@ sub load_metadata {
 
     $self->{parsed_metadata_tags} = 1;
   }
+
+  # see if we found an implicit metadatum for this one; if so, use it.
+  if (!defined $self->{main}->{metadatas}->{$key}
+  	&& defined $self->{implicit_metas}->{$subkey})
+  {
+    $self->add_implicit_metas ($name);
+    return;
+  }
 }
 
 # -------------------------------------------------------------------------
@@ -243,6 +308,8 @@ sub tag_wmmeta {
   my $val = $attrs->{value};
   if (!defined $val) { $val = $text; }
 
+  dbg2 ("adding wmmeta $name: \"$val\"");
+
   $self->{main}->add_metadata ($self->{meta_from}, $name, $val, $attrs,
     		$self->{set_thisdot_metadata_items});
 
@@ -254,9 +321,11 @@ sub tag_wmmeta {
 sub infer_implicit_metas {
   my ($self) = @_;
 
+  dbg2("NormalContent->infer_implicit_metas");
   if (defined $self->{main}->{metadatas}->{"this.title"}
   	&& defined $self->{main}->{metadatas}->{$self->{name}.".title"})
   {
+    dbg2 ("already got titles for this.title and $self->{name}.title");
     return;		# no need to infer it, it's already defined
   }
 
@@ -281,7 +350,7 @@ sub find_implicit_title_in_text {
   elsif ($fmt eq 'text/html') {
     if ($$txt =~ /<title>(.*?)<\/title>/si)
       { $self->add_inferred_metadata ("title", $1, 'text/html'); }
-    # or title tag
+    # or h1/h2/etc. tag
     elsif ($$txt =~ /<h\d>(.*?)<\/h\d>/si)
       { $self->add_inferred_metadata ("title", $1, 'text/html'); }
   }
@@ -296,7 +365,8 @@ sub find_implicit_title_in_text {
   }
 
   # otherwise the first line of non-white chars
-  elsif ($$txt =~ /^\s*(\S[^\n]*?)\s*\n/s)
+  elsif ($fmt =~ /^(text|application)\// &&
+  			$$txt =~ /^\s*(\S[^\n]*?)\s*\n/s)
     { $self->add_inferred_metadata ("title", $1, $fmt); }
 
   undef;
@@ -307,9 +377,6 @@ sub find_implicit_title_in_text {
 sub add_inferred_metadata {
   my ($self, $name, $val, $fmt) = @_;
 
-  my $existingmeta = $self->{main}->{metadatas}->{$self->{name}.".".$name};
-  return if (defined $existingmeta);
-
   my $attrs = { };
 
   if ($fmt ne 'text/html') {
@@ -326,26 +393,45 @@ sub add_inferred_metadata {
   $val =~ s/\s+$//;
 
   dbg ("inferring $name metadata from text: \"$val\"");
-  $self->{main}->add_metadata ($self->{name}, $name, $val, $attrs,
-    		$self->{set_thisdot_metadata_items});
+  
+  #$self->{main}->add_metadata ($self->{name}, $name, $val, $attrs,
+  			#$self->{set_thisdot_metadata_items});
 
-  $self->create_extra_metas_if_needed();
-  $self->{extra_metas}->{$name} = $val;
+  $self->create_implicit_metas_if_needed();
+  $self->{implicit_metas}->{$name} = $val;
 }
 
 # -------------------------------------------------------------------------
 
 sub add_extra_metas {
   my ($self, $from) = @_;
-  # also add our own extra metadata from nav links, <defaultmeta> tags
-  # etc.
+
+  # add our own extra metadata from nav links, <metadefault> tags etc.
   my ($metaname, $val);
   while (($metaname, $val) = each %{$self->{extra_metas}}) {
+    dbg2 ("adding extra meta, from $from,  $metaname: \"$val\"");
     $self->{main}->add_metadata ($from, $metaname, $val, { },
     		$self->{set_thisdot_metadata_items});
   }
 }
 
+sub add_implicit_metas {
+  my ($self, $from) = @_;
+
+  # add our own implicit metadata from <title> tags etc.
+  my ($metaname, $val);
+  while (($metaname, $val) = each %{$self->{implicit_metas}}) {
+    if (defined ($self->{main}->{metadatas}->{$self->{name}.".".$metaname})
+  	&&
+        defined ($self->{main}->{metadatas}->{"this.".$metaname}))
+      { next; }
+
+    dbg2 ("adding implicit meta, from $from,  $metaname: \"$val\"");
+    $self->{main}->add_metadata ($from, $metaname, $val, { },
+    	$self->{set_thisdot_metadata_items});
+  }
+}
+
 # -------------------------------------------------------------------------
 
 sub get_score {
@@ -371,6 +457,18 @@ sub get_modtime {
 
 # -------------------------------------------------------------------------
 
+sub get_edit_href {
+  my ($self) = @_;
+
+  if (defined $self->{datasource}) {
+    return $self->{datasource}->get_location_edit_href ($self->get_filename());
+  } else {
+    return $self->{main}->{cgiglue}->normal_content_in_wmkf_href($self);
+  }
+}
+
+# -------------------------------------------------------------------------
+
 sub get_text_as {
   my ($self, $format) = @_;
   my $main = $self->{main};
@@ -411,6 +509,8 @@ sub get_text_as {
 
   } elsif ($self->{keep_as_is}) {
     $txt = $self->get_as_is_text();
+    dbg ("as-is content $self->{name}: sum=".
+		unpack ("%32C*", $txt)." len=".length($txt));
 
   } else {
     $txt = $self->get_normal_content_text();
@@ -433,7 +533,7 @@ sub get_text_as {
   }
 
   # preformat user tags
-  $main->getusertags()->subst_preformat_tags ($self->{name}, \$txt);
+  1 while ($main->getusertags()->subst_preformat_tags ($self->{name}, \$txt));
 
   # reformat before substs; this way we can cache the reformat
   # results for next time.
@@ -454,6 +554,12 @@ sub get_text_as {
     $txt =~ s/^\s+//s;$txt =~ s/\s+$//s;
   }
 
+  foreach my $hook ($main->get_content_xform_hooks()) {
+    my $cb = $hook->{cb};
+    my $args = $hook->{args};
+    &$cb (@{$args}, $self, \$txt);
+  }
+
   $txt;
 }
 
@@ -620,7 +726,7 @@ sub set_navlinks_vars {
 
   foreach my $dir (qw{up prev next}) {
     my $contname = $self->{main}->curly_meta_subst
-				  ($HTML::WebMake::Main::SUBST_EVAL, "this.nav_".$dir."?");
+			($HTML::WebMake::Main::SUBST_EVAL, "this.nav_".$dir."?");
     my ($obj, $url);
 
     if ($contname ne '') {
@@ -647,9 +753,9 @@ sub set_navlinks_vars {
       $self->{main}->set_transient_content ($dir."text",
 	      $self->navlinks_subst ($self->{'nav_'.$dir}));
 
-    } elsif (defined $self->{'no_nav_'.$dir}) {		# optional attribute
+    } elsif (defined $self->{'nav_no_'.$dir}) {		# optional attribute
       $self->{main}->set_transient_content ($dir."text",
-	      $self->navlinks_subst ($self->{'no_nav_'.$dir}));
+	      $self->navlinks_subst ($self->{'nav_no_'.$dir}));
 
     } else {
       $self->{main}->set_transient_content ($dir."text", '');
