@@ -37,6 +37,12 @@ sub new ($$$$$) {
     }
   }
 
+  # what set of metadata do we use when parsing this out tag?
+  $attrval = $attrs->{'metaset'};
+  $attrval ||= $self->{main}->{metadata}->get_attrdefault ('metaset');
+  $attrval ||= 'def';
+  $self->{metaset} = $attrval;
+
   if (defined $attrs->{root}) {
     $self->{main}->fail ("<out> tags cannot have root attribute: $name");
   }
@@ -56,6 +62,12 @@ sub new ($$$$$) {
   }
   $self->{clean} = $attrval;
 
+  $attrval = $attrs->{'lang'};
+  if (!defined $attrval) {
+    $attrval = $self->{main}->{metadata}->get_attrdefault ('lang');
+  }
+  $self->{lang} = $attrval;
+
   $self;
 }
 
