@@ -41,6 +41,10 @@ package HTML::WebMake::PerlCode;
 use Carp;
 use strict;
 
+use vars qw($DEFER_TAG_EXPANSION);
+
+$DEFER_TAG_EXPANSION = [ -2965179317 ];     # 0xB0BD0BB5 ;)
+
 ###########################################################################
 
 =item $expandedtext = expand ($text);
@@ -194,13 +198,18 @@ sub get_content {
 Get the item of content named, but in Perl list format. It is assumed that the
 list is stored in the content item in whitespace-separated format.
 
+Note that you may have to assign this list to an array, to force it to be
+interpreted by perl as an array instead of as a scalar.   This is annoying,
+but seems unavoidable.
+
 =cut
 
 sub get_list {
   my ($self, $key) = @_;
   if (!defined $key) { croak ("get_list with undef key"); }
   my $str = $self->{main}->curly_or_meta_subst ($HTML::WebMake::Main::SUBST_EVAL, $key);
-  split (' ', $str);
+  my @ret = split (' ', $str);
+  return @ret;
 }
 
 =item set_content ($name, $value);
@@ -530,6 +539,130 @@ sub get_webmake_main_object {
   $self->{main};
 }
 
+# -------------------------------------------------------------------------
+
+=item $filename = get_tmp_filename($type, $extension);
+
+Get a path to a temporary file in the WebMake ~/.webmake directory.
+Useful for plugins.  You should provide a string to use in the filename
+as a clue to the tag type, e.g. "freetable", "thumbnail" etc.; and
+you should provide the extension to use on the file, e.g. "html", "txt",
+"gif" etc.
+
+=cut
+
+sub get_tmp_filename {
+  my ($self, $type, $ext) = @_;
+  return $self->{main}->tmpfile ($type, $ext);
+}
+
+# -------------------------------------------------------------------------
+
+=item $text = scrape_xml ($text, qr/start/, qr/end/ [, $keepstart, $keepend ]);
+
+''Scrape'' a block of HTML or XML text.  Given the text in C<$text>, and
+regular expressions in C<qr/start/> and C<qr/end/>, this function will remove
+all HTML up to and including the start regexp, and all HTML including and
+after the end regexp.
+
+If C<$keepstart> or C<$keepend> is set to 1, then the text matched by that
+regexp will be preserved, otherwise it will be stripped.  The default values
+are 0.
+
+If the patterns match halfway through a HTML or XML tag, then the remainder of
+that tag (until the trailing > character) will be stripped automatically.
+
+If a regexp is specified as C<undef>, then it will be ignored.
+
+The resulting, scraped text is returned.
+
+=cut
+
+sub scrape_xml {
+  my ($self, $text, $startpat, $endpat, $startkeep, $endkeep) = @_;
+  $startkeep ||= 0;
+  $endkeep ||= 0;
+  my $res;
+
+  if (defined $startpat) {
+    if ($startkeep) {
+      $res = ($text =~ s#^.*?(${startpat})#$2#gs);
+    } else {
+      $res = ($text =~ s#^.*?${startpat}##gs);
+    }
+
+    if ($res) { $text =~ s#^[^<>]*?>##s; }
+  }
+
+  if (defined $endpat) {
+    if ($endkeep) {
+      $res = ($text =~ s#(${endpat}).*?$#$1#s);
+    } else {
+      $res = ($text =~ s#${endpat}.*?$##s);
+    }
+
+    if ($res) { $text =~ s#<[^<>]*$##s; }
+  }
+
+  $text;
+}
+
+# -------------------------------------------------------------------------
+
+=item $text = scrape_out_xml ($text, qr/start/, qr/end/ [, $keepstart, $keepend ]);
+
+The inverse of C<scrape_xml()>, above.
+
+Given the text in C<$text>, and regular expressions in C<qr/start/> and
+C<qr/end/>, this function will remove all HTML after, and including, the start
+regexp, and all HTML up to and including the end regexp.
+
+If C<$keepstart> or C<$keepend> is set to 1, then the text matched by that
+regexp will be preserved, otherwise it will be stripped.  The default values
+are 0.
+
+If the patterns match halfway through a HTML or XML tag, then the remainder of
+that tag (until the trailing > character) will be stripped automatically.
+
+The regexps cannot be specified as C<undef>, as C<scrape_xml()> should
+be used for that case instead.
+
+The resulting, scraped text is returned.
+
+=cut
+
+sub scrape_out_xml {
+  my ($self, $text, $startpat, $endpat, $startkeep, $endkeep) = @_;
+  $startkeep ||= 0;
+  $endkeep ||= 0;
+  my $res;
+
+  # we use this token to match the bit we scoop out
+  my $scrapedout = '  <!!!\002>'.'__SCRAPED_OUT__'.'<\002!!!>  ';
+
+  if ($startkeep) {
+    if ($endkeep) {
+      $res = ($text =~ s#(${startpat}).*?(${endpat})#$1 ${scrapedout} $2#s);
+    } else {
+      $res = ($text =~ s#(${startpat}).*?${endpat}#$1 ${scrapedout} #s);
+    }
+  } else {
+    if ($endkeep) {
+      $res = ($text =~ s#${startpat}.*?(${endpat})# ${scrapedout} $1#s);
+    } else {
+      $res = ($text =~ s#${startpat}.*?${endpat}# ${scrapedout} #s);
+    }
+  }
+
+  if ($res) {
+    $text =~ s#${scrapedout}[^<>]*?>#${scrapedout}#so;
+    $text =~ s#<[^<>]*${scrapedout}##so;
+    $text =~ s#${scrapedout}##so;
+  }
+
+  $text;
+}
+
 # -------------------------------------------------------------------------
 # Glue functions -- these allow calls from perl code without using the
 # $self->function_name() OO mode.
@@ -585,6 +718,12 @@ sub get_current_main_content
 { $HTML::WebMake::PerlCode::GlobalSelf->get_current_main_content(@_); }
 sub get_webmake_main_object
 { $HTML::WebMake::PerlCode::GlobalSelf->get_webmake_main_object(@_); }
+sub get_tmp_filename
+{ $HTML::WebMake::PerlCode::GlobalSelf->get_tmp_filename(@_); }
+sub scrape_xml
+{ $HTML::WebMake::PerlCode::GlobalSelf->scrape_xml(@_); }
+sub scrape_out_xml
+{ $HTML::WebMake::PerlCode::GlobalSelf->scrape_out_xml(@_); }
 sub expand
 { $HTML::WebMake::PerlCode::GlobalSelf->expand(@_); }
 
