@@ -19,10 +19,10 @@ use vars	qw{
 
 ###########################################################################
 
-sub new ($$$$$) {
+sub new ($$$$$$$$) {
   my $class = shift;
   $class = ref($class) || $class;
-  my ($main, $filename, $outname, $dotdots, $fmt, $useurls) = @_;
+  my ($main, $filename, $outname, $dotdots, $fmt, $useurls, $metaset, $lang) = @_;
 
   my $self = {
     'main'		=> $main,
@@ -33,6 +33,8 @@ sub new ($$$$$) {
     'dotdots'		=> $dotdots,
     'useurls'		=> $useurls,
     'format'		=> $fmt,
+    'metaset'		=> $metaset,
+    'lang'			=> $lang
   };
   bless ($self, $class);
 
