@@ -208,5 +208,25 @@ sub text_eol {
 
 ###########################################################################
 
+sub timespec_to_seconds {
+  my ($self, $tspec) = @_;
+  my $secs = 0;
+
+  $tspec =~ s/\s+/ /g; $tspec = " $tspec ";
+  if ($tspec =~ s/ (\d+) *d\S* / /i) {
+    $secs += $1 * (24*60*60);
+  }
+  if ($tspec =~ s/ (\d+) *h\S* / /i) {
+    $secs += $1 * (60*60);
+  }
+  if ($tspec =~ s/ (\d+) *m\S* / /i) {
+    $secs += $1 * (60);
+  }
+  if ($tspec =~ s/ (\d+) / /i) {
+    $secs += $1;
+  }
+  return $secs;
+}
+
 1;
 
