@@ -34,6 +34,10 @@ sub new ($$$) {
   $self->{cgi} = {
     'fulltext'		=> undef,
     'items'		=> [ ],
+    keep_cgi_fulltext	=> 0,
+    skip_eval_code	=> 0,
+    skip_subst_attrs	=> 0,
+    mk_cgi_items_array	=> 0,
   };
 
   bless ($self, $class);
@@ -53,7 +57,7 @@ sub parse {
 
   if (!defined $self->{main}) { carp "no main defined in WmkFile::parse"; }
 
-  if ($self->{parse_for_cgi}) {
+  if ($self->{keep_cgi_fulltext}) {
     $self->{cgi}->{fulltext} = $_;
   }
 
@@ -69,37 +73,11 @@ sub parse {
   s/^.*?<webmake\b[^>]*?>//gis;
   s/<\/\s*webmake\s*>.*$//gis;
 
-  # handle scoped tags.  Since we don't use a proper XML parser, we have to
-  # rewrite them here.  We convert them to single-character markers (\001 or
-  # \002) indicating a start tag or end tag, then loop until all appearances of
-  # the tag have been converted. We then convert them back to text, with a
-  # scope number attached.  Until Perl can do a regexp like this:
-  #
-  # /<tag[^>]*>[^<tag]+<\/tag>/
-  #
-  # we're probably stuck doing it this way.  Hey, don't knock it, it works ;)
-
-  s/\001/<<001>>/gs;
-  s/\002/<<002>>/gs;
   $self->{scopings} = { };
-  for my $tag (qw(for metadefault attrdefault)) {
-    if (!/<\/$tag>/) {
-      $self->{scopings}->{$tag} = 0; next;
-    }
-
-    s/<$tag(\b[^>]*[^\/]>)/\001$1/gs;
-    s/<\/$tag>/\002/gs;
-
-    my $count = 0;
-    while (s{\001([^>]+)>([^\001\002]+)\002}
-    		{<$tag$count$1>$2<\/$tag$count>}gis)
-    {
-      $count++;
-    }
-    $self->{scopings}->{$tag} = $count;
+  for my $tag (qw(for metadefault attrdefault usemetaset)) {
+    $self->{scopings}->{$tag} = 0;
   }
-  s/<<001>>/\001/gs;
-  s/<<002>>/\002/gs;
+  $self->fix_scoped_tags (\$_);
 
   my $util = $self->{main}->{util};
   if (!defined $util) { carp "no util defined in WmkFile::parse"; }
@@ -109,7 +87,7 @@ sub parse {
   # if we are parsing for the CGI scripts, make sure that the XML
   # parser also notes regular expressions which match each item, so that the
   # CGI code can rewrite the file easily later.
-  if ($self->{parse_for_cgi}) {
+  if ($self->{mk_cgi_items_array}) {
     $util->{generate_tag_regexps} = 1;
   }
 
@@ -126,13 +104,17 @@ sub parse {
     1 while s/^<!--.*?-->//gs;		# XML-style comments.
 
     # Preprocessing.
-    $util->strip_first_lone_tag (\$_, "include",
-				  $self, \&tag_include, qw(file));
-    $util->strip_first_lone_tag (\$_, "use",
-				  $self, \&tag_use, qw(plugin));
-
-    if (!$self->{parse_for_cgi}) {
+    if (!$self->{skip_eval_code}) {
       $self->{main}->eval_code_at_parse (\$_);
+
+      # This may have been set by a plugin that needs introspection
+      # into the .wmk file; check it here.
+      if ($self->{main}->{start_parsing_introspection}) {
+	$self->set_parse_introspection();
+	$util->{generate_tag_regexps} = 1;
+        $self->{main}->{start_parsing_introspection} = 0;
+      }
+
     } else {
       1 while s/^<{.*?}>//gs;		# trim code, CGI mode doesn't need it
     }
@@ -146,19 +128,25 @@ sub parse {
       if (defined $text) { $lasteval = $text; $lasttag = undef; }
     }
 
-    # Declarations.
+    # Tags, ordered from most-likely to least-likely...
+    $util->strip_first_tag_block (\$_, "out",
+				  $self, \&tag_out, qw(file));
     $util->strip_first_tag_block (\$_, "content",
 				  $self, \&tag_content, qw(name));
-    $util->strip_first_lone_tag (\$_, "contents",
-				  $self, \&tag_contents, qw(src name));
     $util->strip_first_tag_block (\$_, "template",
 				  $self, \&tag_template, qw(name));
+    $util->strip_first_lone_tag (\$_, "include",
+				  $self, \&tag_include, qw(file));
+    $util->strip_first_lone_tag (\$_, "use",
+				  $self, \&tag_use, qw(plugin));
+    $util->strip_first_lone_tag (\$_, "contents",
+				  $self, \&tag_contents, qw(src name));
     $util->strip_first_lone_tag (\$_, "templates",
 				  $self, \&tag_templates, qw(src name));
-    $util->strip_first_tag_block (\$_, "contenttable",
-				  $self, \&tag_contenttable, qw());
     $util->strip_first_lone_tag (\$_, "media",
 				  $self, \&tag_media, qw(src name));
+    $util->strip_first_tag_block (\$_, "contenttable",
+				  $self, \&tag_contenttable, qw());
 
     if (/^<metadefault/i) {
       $util->strip_first_lone_tag (\$_, "metadefault",
@@ -178,9 +166,18 @@ sub parse {
 				    $self, \&tag_attrdefault, qw(name));
       }
     }
+    if (/^<usemetaset/i) {
+      my $i;
+      for ($i = 0; $i < $self->{scopings}->{"usemetaset"}; $i++) {
+	$util->strip_first_tag_block (\$_, "usemetaset".$i,
+				    $self, \&tag_usemetaset, qw(id));
+      }
+    }
 
     $util->strip_first_tag (\$_, "metatable",
 				  $self, \&tag_metatable, qw());
+    $util->strip_first_tag (\$_, "metaset",
+				  $self, \&tag_metaset, qw(id));
     $util->strip_first_tag (\$_, "sitemap",
 				  $self, \&tag_sitemap, qw(name node leaf));
     $util->strip_first_tag (\$_, "navlinks",
@@ -199,10 +196,6 @@ sub parse {
       }
     }
 
-    # Outputs.
-    $util->strip_first_tag_block (\$_, "out",
-				  $self, \&tag_out, qw(file));
-
     # Misc.
     $util->strip_first_lone_tag (\$_, "cache",
 				  $self, \&tag_cache, qw(dir));
@@ -216,6 +209,8 @@ sub parse {
 				  $self, \&tag_viewcgi, qw(href));
     $util->strip_first_lone_tag (\$_, "site",
 				  $self, \&tag_site, qw(href));
+    $util->strip_first_tag_block (\$_, "action",
+				  $self, \&tag_action, qw(event));
 
     # if we got some tags, store the text for error messages
     my $text = $util->{last_tag_text};
@@ -241,7 +236,8 @@ sub parse {
       $failuretext = $lasteval;
     }
 
-    /^(.*?>.{40,40})/s; $_ = $1; $_ =~ s/\s+/ /gs;
+    /^([^<].*?>.{40,40})/s; if (defined $1) { $_ = $1; }
+    s/\s+/ /gs;
     $lasttag ||= '';
     $self->{main}->fail ("WMK file contains unparseable data at or after:\n".
 	      "\t$lasttag\n\t$_ ...\"\n");
@@ -253,9 +249,45 @@ sub parse {
 
 # -------------------------------------------------------------------------
 
+# handle scoped tags.  Since we don't use a proper XML parser, we have to
+# rewrite them here.  We convert them to single-character markers (\001 or
+# \002) indicating a start tag or end tag, then loop until all appearances of
+# the tag have been converted. We then convert them back to text, with a
+# scope number attached.  Until Perl can do a regexp like this:
+#
+# /<tag[^>]*>[^<tag]+<\/tag>/
+#
+# we're probably stuck doing it this way.  Hey, don't knock it, it works ;)
+
+sub fix_scoped_tags {
+  my ($self, $txt) = @_;
+
+  $$txt =~ s/\001/<<001>>/gs;
+  $$txt =~ s/\002/<<002>>/gs;
+
+  for my $tag (qw(for metadefault attrdefault usemetaset)) {
+    next if ($$txt !~ /<\/$tag>/);
+
+    $$txt =~ s/<$tag(\b[^>]*[^\/]>)/\001$1/gs;
+    $$txt =~ s/<\/$tag>/\002/gs;
+
+    my $count = $self->{scopings}->{$tag};
+    while ($$txt =~ s{\001([^>]+)>([^\001\002]+)\002}
+    			{<$tag$count$1>$2<\/$tag$count>}gis)
+    {
+      $self->{scopings}->{$tag}++;
+      $count++;
+    }
+  }
+  $$txt =~ s/<<001>>/\001/gs;
+  $$txt =~ s/<<002>>/\002/gs;
+}
+
+# -------------------------------------------------------------------------
+
 sub subst_attrs {
   my ($self, $tagname, $attrs) = @_;
-  return if ($self->{parse_for_cgi});
+  return if ($self->{skip_subst_attrs});
 
   if (defined ($attrs->{name})) {
     $tagname .= " \"".$attrs->{name}."\"";	# for errors
@@ -288,6 +320,8 @@ sub tag_include {
   dbg ("included file: \"$file\"");
   $self->{main}->set_file_modtime ($file, $s[9]);
   $self->add_dep ($file);
+
+  $self->fix_scoped_tags (\$inc);
   $inc;
 }
 
@@ -317,7 +351,6 @@ sub tag_use {
   }
 
 foundit:
-
   if (!open (INC, "<$file")) {
     die "Cannot open 'use' file: $file\n";
   }
@@ -327,7 +360,19 @@ foundit:
   dbg ("used file: \"$file\"");
   $self->{main}->set_file_modtime ($file, $s[9]);
   $self->add_dep ($file);
-  $inc;
+
+  $self->fix_scoped_tags (\$inc);
+
+  # make the <use> tag's attributes available to the plugin
+  my $attrstr = '';
+  foreach my $key (keys %{$attrs}) {
+    next if ($key eq 'plugin');
+    my $val = $attrs->{$key};
+    $val =~ s!</template>!\&lt;/template>!g;	# just in case
+    $attrstr .= "<template name=\"$plugin.$key\">$val</template>\n";
+  }
+
+  $attrstr.$inc;
 }
 
 # -------------------------------------------------------------------------
@@ -357,7 +402,10 @@ sub tag_editcgi {
   my ($self, $tag, $attrs, $text) = @_;
 
   $self->subst_attrs ("<editcgi>", $attrs);
-  $self->{main}->add_url ("WebMake.EditCGI", $attrs->{href});
+  my $wmkf = 'main.wmk';		# TODO
+
+  $self->{main}->add_url ("WebMake.EditCGI", $attrs->{href}
+  		."?wmkf=".$wmkf);
   "";
 }
 
@@ -373,6 +421,16 @@ sub tag_viewcgi {
 
 # -------------------------------------------------------------------------
 
+sub tag_action {
+  my ($self, $tag, $attrs, $text) = @_;
+
+  $self->subst_attrs ("<action>", $attrs);
+  $self->{main}->add_action ($attrs->{event}, $text);
+  "";
+}
+
+# -------------------------------------------------------------------------
+
 sub tag_site {
   my ($self, $tag, $attrs, $text) = @_;
 
@@ -400,6 +458,10 @@ sub tag_content {
     $attrs->{isroot} = $attrs->{root};	# backwards compat
   }
 
+  if (defined $attrs->{src}) {
+    $text = $self->read_src ('content', $attrs, $text);
+  }
+
   $self->{main}->add_content ($name, $self, $attrs, $text);
   "";
 }
@@ -427,6 +489,10 @@ sub tag_template {
   }
   $attrs->{map} = 'false';
 
+  if (defined $attrs->{src}) {
+    $text = $self->read_src ('template', $attrs, $text);
+  }
+
   $self->{main}->add_content ($name, $self, $attrs, $text);
   "";
 }
@@ -460,6 +526,10 @@ sub tag_contenttable {
   $self->cgi_add ($tag, $CGI_EDIT_AS_TEXT, $text, $attrs) and return '';
   $self->subst_attrs ("<contenttable>", $attrs);
 
+  if (defined $attrs->{src}) {
+    $text = $self->read_src ('contenttable', $attrs, $text);
+  }
+
   # we actually use a Contents object, reading from the .wmk file
   # to do this.
   $attrs->{src} = 'svfile:';
@@ -499,6 +569,32 @@ sub tag_attrdefault {
   $text . '<attrdefault name="'.$attrs->{name}.'" value="[POP]" />';
 }
 
+sub tag_usemetaset {
+  my ($self, $tag, $attrs, $text) = @_;
+
+  $self->cgi_add ($tag, $CGI_NON_EDITABLE, undef, $attrs) and return $text;
+  $self->subst_attrs ("<usemetaset>", $attrs);
+
+  # set it on Main in case perl code accesses metadata (quite likely!)
+  $self->{main}->{metaset} = $attrs->{id};
+
+  # also set it on Content items etc.; sneaky kludge!
+  # Use 'attrdefault' to implement this
+  $self->{main}->{metadata}->set_attrdefault ('usemetaset', $attrs->{id});
+
+  return '' if (!defined $text || $text eq '');
+  $text . '<attrdefault name="usemetaset" value="[POP]" />';
+}
+
+sub tag_metaset {
+  my ($self, $tag, $attrs, $text) = @_;
+
+  $self->cgi_add ($tag, $CGI_NON_EDITABLE, undef, $attrs) and return $text;
+  $self->subst_attrs ("<metaset>", $attrs);
+  $self->{main}->{metadata}->parse_metaset ($attrs->{id}, $text, $attrs);
+  '';
+}
+
 sub tag_metatable {
   my ($self, $tag, $attrs, $text) = @_;
 
@@ -506,14 +602,7 @@ sub tag_metatable {
   $self->subst_attrs ("<metatable>", $attrs);
 
   if (defined $attrs->{src}) {
-    my $fname = $attrs->{src};
-    if (open (IN, "<".$fname)) {
-      $text = join ('', <IN>);
-      close IN;
-
-    } else {
-      warn ("<metatable src=\"$attrs->{src}\"> could not be read: $@\n");
-    }
+    $text = $self->read_src ('metatable', $attrs, $text);
   }
 
   my $tbl = new HTML::WebMake::MetaTable ($self->{main});
@@ -560,7 +649,12 @@ sub tag_out {
   $self->subst_attrs ("<out>", $attrs);
   my $file = $attrs->{file};
   my $name = $attrs->{name}; $name ||= $file;
+
   $self->{main}->add_out ($file, $self, $name, $attrs, $text);
+
+  my $lang = $self->{main}->{outs}->{$file}->{lang};
+  $name = $name."::".$lang if defined $lang;
+
   $self->{main}->add_url ($name, $file);
   "";
 }
@@ -581,9 +675,11 @@ sub tag_for ($$$$) {
   {
     if (!$self->{main}->{paranoid}) {
       if (defined $namesubst) {
-	@vals = map { eval $namesubst; $_; } @vals;
+	eval '
+	  @vals = map { '.$namesubst.'; $_; } @vals;
+	';
       }
-      if ($#vals < 0) {
+      if ($#vals < 0 || $@) {
 	warn ("<for> tag \"$attrs->{name}\" namesubst failed: $@\n");
       }
 
@@ -608,7 +704,7 @@ sub tag_for ($$$$) {
 sub cgi_add {
   my ($self, $tag, $editui, $edituidata, $attrs) = @_;
 
-  return undef unless ($self->{parse_for_cgi});
+  return undef unless ($self->{mk_cgi_items_array});
 
   my $name = "$tag";
   if (defined $attrs->{name}) {
@@ -634,13 +730,15 @@ sub cgi_add {
   };
 
   push (@{$self->{cgi}->{items}}, $item);
+
+  if (!$self->{skip_adding_items}) { return undef; }
   return ' ';
 }
 
 sub cgi_add_datasource {
   my ($self, $tag, $attrs) = @_;
 
-  return undef unless ($self->{parse_for_cgi});
+  return undef unless ($self->{mk_cgi_items_array});
 
   my $proto = 'file';
   my $src = $attrs->{src};
@@ -652,9 +750,98 @@ sub cgi_add_datasource {
     $self->cgi_add ($tag, $CGI_EDIT_AS_DIR, $src, $attrs);
   }
 
+  if (!$self->{skip_adding_items}) { return undef; }
   return ' ';
 }
 
 ###########################################################################
 
+sub read_src {
+  my ($self, $tag, $attrs, $oldtext) = @_;
+
+  my $src = $attrs->{src};
+  my $text;
+  my $file;
+
+  if ($src =~ /^(http|https|ftp):/) {
+    my $updfreq = $attrs->{updatefreq};
+    $updfreq ||= "1h";			# 1 hour by default
+    $updfreq = $self->{main}->{util}->timespec_to_seconds ($updfreq);
+
+    dbg ("attempting to load \"$src\" using LWP...");
+
+    # using LWP::Simple::mirror, we can cache the content to disk
+    # and only refetch it when it changes.
+    #
+    my $tmpdir = File::Spec->catdir ($self->{main}->tmpdir(), "srcs");
+    if (! -d $tmpdir) { mkdir ($tmpdir, 0777); }
+
+    my $tmpfile = $src;
+    $tmpfile =~ s/[^-_=\.\,\+A-Za-z0-9]+/_/gs;
+    $tmpfile = File::Spec->catdir ($tmpdir, $tmpfile);
+
+    my $rcode;
+
+    if (-f $tmpfile && (-M $tmpfile) * 24*60*60 < $updfreq) {
+      dbg ("cached copy is young enough, not refetching");
+      $file = $tmpfile;
+
+    } else {
+      eval {
+	  require LWP::Simple;
+	  $rcode = LWP::Simple::mirror ($src, $tmpfile);
+      };
+
+      if ($@) {
+	warn "HTTP get $src failed: LWP module not available\n";
+      } elsif (LWP::Simple::is_error ($rcode)) {
+	warn "HTTP get $src failed: HTTP error code $rcode.\n";
+      } else {
+	$file = $tmpfile;
+      }
+    }
+
+  } else {
+    $file = $src;
+  }
+
+  if (defined $file) {
+    $file =~ s/^file:\/+/\//g;	# remove optional file: prefix
+    if (open (IN, "<".$file)) {
+      binmode IN;		# in case it's an image etc.
+      $text = '';
+      while (sysread IN, $_, 4096) { $text .= $_; }
+      close IN;
+
+      dbg ("$file: sum=".unpack ("%32C*", $text)." len=".length($text));
+    }
+  }
+
+  if (!defined $text) {
+    warn ("<$tag src=\"$attrs->{src}\"> could not be read.\n");
+    return '';
+  }
+
+  return $text;
+}
+
+###########################################################################
+
+# CGI-only introspection: inhibits code parsing and attr substs
+sub set_parse_for_cgi {
+  my $self = shift;
+  $self->{keep_cgi_fulltext} = 1;
+  $self->{mk_cgi_items_array} = 1;
+  $self->{skip_adding_items} = 1;
+  $self->{skip_eval_code} = 1;
+  $self->{skip_subst_attrs} = 1;
+}
+
+# lightweight introspection: do not inhibit normal parsing.
+sub set_parse_introspection {
+  my $self = shift;
+  $self->{keep_cgi_fulltext} = 1;
+  $self->{mk_cgi_items_array} = 1;
+}
+
 1;
