@@ -112,7 +112,7 @@ sub patterns_run_cb {
     my $safe = pattern_to_re ($pat);
     # print "JMD $patterns{$pat}\n";
     if ($_ =~ /${safe}/s) {
-      $found_anti{$patterns{$pat}}++;
+      $found_anti{$anti_patterns{$pat}}++;
     }
   }
 }
@@ -121,8 +121,14 @@ sub ok_all_patterns {
   foreach my $pat (sort keys %patterns) {
     my $type = $patterns{$pat};
     print "\tChecking $type\n";
+
     if (ok (defined $found{$type})) {
-      ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      if ($allowed_multi{$type}) {
+	ok (1);
+      } else {
+	ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      }
+
     } else {
       warn "\tNot found: $type = $pat\n";
       ok (0);                     # keep the right # of tests
